@@ -11,47 +11,45 @@ and conditions of the chosen license apply to this file.
 #[cfg(feature = "image-data")]
 use crate::common::ImageData;
 use crate::common::{private, Error};
-use std::{borrow::Cow, marker::PhantomData, path::PathBuf, thread, time::Duration};
+use std::{
+	borrow::Cow,
+	io,
+	marker::PhantomData,
+	os::windows::{fs::OpenOptionsExt, io::AsRawHandle},
+	path::{Path, PathBuf},
+	thread,
+	time::Duration,
+};
+use windows_sys::Win32::{
+	Foundation::{GetLastError, GlobalFree, HANDLE, HGLOBAL, POINT, S_OK},
+	Storage::FileSystem::{GetFinalPathNameByHandleW, FILE_FLAG_BACKUP_SEMANTICS, VOLUME_NAME_DOS},
+	System::{
+		DataExchange::SetClipboardData,
+		Memory::{GlobalAlloc, GlobalLock, GlobalUnlock, GHND},
+		Ole::CF_HDROP,
+	},
+	UI::Shell::{PathCchStripPrefix, DROPFILES},
+};
 
 #[cfg(feature = "image-data")]
 mod image_data {
 	use super::*;
 	use crate::common::ScopeGuard;
+	use image::codecs::bmp::BmpDecoder;
+	use image::codecs::png::PngDecoder;
 	use image::codecs::png::PngEncoder;
+	use image::DynamicImage;
 	use image::ExtendedColorType;
+	use image::ImageDecoder;
 	use image::ImageEncoder;
-	use std::{convert::TryInto, io, mem::size_of, ptr::copy_nonoverlapping};
+	use std::{convert::TryInto, mem::size_of, ptr::copy_nonoverlapping};
 	use windows_sys::Win32::{
-		Foundation::{HANDLE, HGLOBAL},
 		Graphics::Gdi::{
-			CreateDIBitmap, DeleteObject, GetDC, GetDIBits, BITMAPINFO, BITMAPINFOHEADER,
-			BITMAPV5HEADER, BI_BITFIELDS, BI_RGB, CBM_INIT, DIB_RGB_COLORS, HBITMAP, HDC, HGDIOBJ,
-			LCS_GM_IMAGES, RGBQUAD,
-		},
-		System::{
-			DataExchange::SetClipboardData,
-			Memory::{GlobalAlloc, GlobalLock, GlobalUnlock, GHND},
-			Ole::CF_DIBV5,
+			DeleteObject, BITMAPV5HEADER, BI_BITFIELDS, BI_RGB, HGDIOBJ, LCS_GM_IMAGES,
 		},
+		System::Ole::CF_DIBV5,
 	};
 
-	fn last_error(message: &str) -> Error {
-		let os_error = io::Error::last_os_error();
-		Error::unknown(format!("{}: {}", message, os_error))
-	}
-
-	unsafe fn global_unlock_checked(hdata: HGLOBAL) {
-		// If the memory object is unlocked after decrementing the lock count, the function
-		// returns zero and GetLastError returns NO_ERROR. If it fails, the return value is
-		// zero and GetLastError returns a value other than NO_ERROR.
-		if GlobalUnlock(hdata) == 0 {
-			let err = io::Error::last_os_error();
-			if err.raw_os_error() != Some(0) {
-				log::error!("Failed calling GlobalUnlock when writing data: {}", err);
-			}
-		}
-	}
-
 	pub(super) fn add_cf_dibv5(
 		_open_clipboard: OpenClipboard,
 		image: ImageData,
@@ -168,182 +166,67 @@ mod image_data {
 		}
 	}
 
-	unsafe fn global_alloc(bytes: usize) -> Result<HGLOBAL, Error> {
-		let hdata = GlobalAlloc(GHND, bytes);
-		if hdata.is_null() {
-			Err(last_error("Could not allocate global memory object"))
-		} else {
-			Ok(hdata)
+	// https://learn.microsoft.com/en-us/windows/win32/api/wingdi/ns-wingdi-bitmapv5header
+	// According to the docs, when bV5Compression is BI_RGB, "the high byte in each DWORD
+	// is not used".
+	// This seems to not be respected in the real world. For example, Chrome, and Chromium
+	// & Electron-based programs send us BI_RGB headers, but with bitCount=32 - and important
+	// transparency bytes in the alpha channel.
+	//
+	// Apparently, it's our job as the consumer to do the right thing. This method fiddles
+	// with the header a bit in these cases, then `image` handles the rest.
+	fn maybe_tweak_header(dibv5: &mut [u8]) {
+		assert!(dibv5.len() >= size_of::<BITMAPV5HEADER>());
+		let src = dibv5.as_mut_ptr().cast::<BITMAPV5HEADER>();
+		let mut header = unsafe { std::ptr::read_unaligned(src) };
+
+		if header.bV5BitCount == 32
+			&& header.bV5Compression == BI_RGB
+			&& header.bV5AlphaMask == 0xff000000
+		{
+			header.bV5Compression = BI_BITFIELDS;
+			if header.bV5RedMask == 0 && header.bV5GreenMask == 0 && header.bV5BlueMask == 0 {
+				header.bV5RedMask = 0xff0000;
+				header.bV5GreenMask = 0xff00;
+				header.bV5BlueMask = 0xff;
+			}
+
+			unsafe { std::ptr::write_unaligned(src, header) };
 		}
 	}
 
-	unsafe fn global_lock(hmem: HGLOBAL) -> Result<*mut u8, Error> {
-		let data_ptr = GlobalLock(hmem).cast::<u8>();
-		if data_ptr.is_null() {
-			Err(last_error("Could not lock the global memory object"))
-		} else {
-			Ok(data_ptr)
-		}
-	}
-
-	pub(super) fn read_cf_dibv5(dibv5: &[u8]) -> Result<ImageData<'static>, Error> {
+	pub(super) fn read_cf_dibv5(dibv5: &mut [u8]) -> Result<ImageData<'static>, Error> {
 		// The DIBV5 format is a BITMAPV5HEADER followed by the pixel data according to
 		// https://docs.microsoft.com/en-us/windows/win32/dataxchg/standard-clipboard-formats
 
-		// These constants are missing in windows-rs
-		const PROFILE_EMBEDDED: u32 = 0x4D42_4544;
-		const PROFILE_LINKED: u32 = 0x4C49_4E4B;
-
-		// so first let's get a pointer to the header
 		let header_size = size_of::<BITMAPV5HEADER>();
 		if dibv5.len() < header_size {
 			return Err(Error::unknown("When reading the DIBV5 data, it contained fewer bytes than the BITMAPV5HEADER size. This is invalid."));
 		}
-		let header = unsafe { &*(dibv5.as_ptr().cast::<BITMAPV5HEADER>()) };
+		maybe_tweak_header(dibv5);
 
-		let has_profile =
-			header.bV5CSType == PROFILE_LINKED || header.bV5CSType == PROFILE_EMBEDDED;
+		let decoder = BmpDecoder::new_without_file_header(std::io::Cursor::new(&*dibv5))
+			.map_err(|_| Error::ConversionFailure)?;
+		let (width, height) = decoder.dimensions();
+		let bytes = DynamicImage::from_decoder(decoder)
+			.map_err(|_| Error::ConversionFailure)?
+			.into_rgba8()
+			.into_raw();
 
-		let pixel_data_start = if has_profile {
-			header.bV5ProfileData as isize + header.bV5ProfileSize as isize
-		} else {
-			header_size as isize
-		};
-
-		unsafe {
-			let image_bytes = dibv5.as_ptr().offset(pixel_data_start);
-			let hdc = get_screen_device_context()?;
-			let hbitmap = create_bitmap_from_dib(hdc, header, image_bytes)?;
-			// Now extract the pixels in a desired format
-			let w = header.bV5Width;
-			let h = header.bV5Height.abs();
-			let result_size = w as usize * h as usize * 4;
-
-			let mut result_bytes = Vec::<u8>::with_capacity(result_size);
-
-			let mut output_header = BITMAPINFO {
-				bmiColors: [RGBQUAD { rgbRed: 0, rgbGreen: 0, rgbBlue: 0, rgbReserved: 0 }],
-				bmiHeader: BITMAPINFOHEADER {
-					biSize: size_of::<BITMAPINFOHEADER>() as u32,
-					biWidth: w,
-					biHeight: -h,
-					biBitCount: 32,
-					biPlanes: 1,
-					biCompression: BI_RGB,
-					biSizeImage: 0,
-					biXPelsPerMeter: 0,
-					biYPelsPerMeter: 0,
-					biClrUsed: 0,
-					biClrImportant: 0,
-				},
-			};
-
-			let lines = convert_bitmap_to_rgb(
-				hdc,
-				hbitmap,
-				h,
-				result_bytes.as_mut_slice(),
-				&mut output_header,
-			)?;
-			let read_len = lines as usize * w as usize * 4;
-			assert!(
-				read_len <= result_bytes.capacity(),
-				"Segmentation fault. Read more bytes than allocated to pixel buffer",
-			);
-			result_bytes.set_len(read_len);
-
-			let result_bytes = win_to_rgba(&mut result_bytes);
-
-			let result = ImageData {
-				bytes: Cow::Owned(result_bytes),
-				width: w as usize,
-				height: h as usize,
-			};
-			Ok(result)
-		}
+		Ok(ImageData { width: width as usize, height: height as usize, bytes: bytes.into() })
 	}
 
-	fn get_screen_device_context() -> Result<HDC, Error> {
-		// SAFETY: Calling `GetDC` with `NULL` is safe.
-		let hdc = unsafe { GetDC(ResultValue::NULL) };
-		if hdc.failure() {
-			Err(Error::unknown("Failed to get the device context. GetDC returned null"))
-		} else {
-			Ok(hdc)
-		}
-	}
+	pub(super) fn read_png(data: &[u8]) -> Result<ImageData<'static>, Error> {
+		let decoder =
+			PngDecoder::new(std::io::Cursor::new(data)).map_err(|_| Error::ConversionFailure)?;
+		let (width, height) = decoder.dimensions();
 
-	unsafe fn create_bitmap_from_dib(
-		hdc: HDC,
-		header: *const BITMAPV5HEADER,
-		image_bytes: *const u8,
-	) -> Result<HBITMAP, Error> {
-		let hbitmap = CreateDIBitmap(
-			hdc,
-			header.cast(),
-			CBM_INIT as u32,
-			image_bytes.cast(),
-			header.cast(),
-			DIB_RGB_COLORS,
-		);
-		if hbitmap.failure() {
-			Err(Error::unknown(
-				"Failed to create the HBITMAP while reading DIBV5. CreateDIBitmap returned null",
-			))
-		} else {
-			Ok(hbitmap)
-		}
-	}
+		let bytes = DynamicImage::from_decoder(decoder)
+			.map_err(|_| Error::ConversionFailure)?
+			.into_rgba8()
+			.into_raw();
 
-	/// Copies the bitmap image into given buffer with DIB RGB format and
-	/// returns the number of scan lines copied from the bitmap.
-	unsafe fn convert_bitmap_to_rgb(
-		hdc: HDC,
-		hbitmap: HBITMAP,
-		lines: i32,
-		dst: &mut [u8],
-		header: &mut BITMAPINFO,
-	) -> Result<i32, Error> {
-		let lines = GetDIBits(
-			hdc,
-			hbitmap,
-			0,
-			lines as u32,
-			dst.as_mut_ptr().cast(),
-			header,
-			DIB_RGB_COLORS,
-		);
-		if lines == 0 {
-			Err(Error::unknown("Could not get the bitmap bits, GetDIBits returned 0"))
-		} else {
-			Ok(lines)
-		}
-	}
-
-	/// An abstraction trait over the different ways a Win32 function may return
-	/// a value with a failure marker.
-	///
-	/// This is primarily to abstract over changes in `windows-sys` versions and unify how
-	/// error handling is done in the above image code.
-	trait ResultValue: Sized {
-		const NULL: Self;
-		fn failure(self) -> bool;
-	}
-
-	// windows-sys >= 0.59
-	impl<T> ResultValue for *mut T {
-		const NULL: Self = core::ptr::null_mut();
-		fn failure(self) -> bool {
-			self == Self::NULL
-		}
-	}
-
-	// `windows-sys` 0.52
-	impl ResultValue for isize {
-		const NULL: Self = 0;
-		fn failure(self) -> bool {
-			self == Self::NULL
-		}
+		Ok(ImageData { width: width as usize, height: height as usize, bytes: bytes.into() })
 	}
 
 	/// Converts the RGBA (u8) pixel data into the bitmap-native ARGB (u32)
@@ -411,6 +294,7 @@ mod image_data {
 	/// Safety: the `bytes` slice must have a length that's a multiple of 4
 	#[allow(clippy::identity_op, clippy::erasing_op)]
 	#[must_use]
+	#[cfg(test)]
 	unsafe fn win_to_rgba(bytes: &mut [u8]) -> Vec<u8> {
 		// Check safety invariants to catch obvious bugs.
 		debug_assert_eq!(bytes.len() % 4, 0);
@@ -485,6 +369,141 @@ mod image_data {
 		let _converted = unsafe { win_to_rgba(&mut data) };
 		assert_eq!(data, DATA);
 	}
+
+	#[test]
+	fn firefox_dibv5() {
+		// A 5x5 sample of https://commons.wikimedia.org/wiki/File:PNG_transparency_demonstration_1.png
+		let mut raw = vec![
+			124, 0, 0, 0, 5, 0, 0, 0, 5, 0, 0, 0, 1, 0, 24, 0, 0, 0, 0, 0, 80, 0, 0, 0, 0, 0, 0, 0,
+			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 255, 0, 0, 0, 0, 255, 0, 0, 0, 0, 255, 0, 0, 0, 0,
+			255, 66, 71, 82, 115, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 36, 47, 144, 42, 68, 110, 48, 74, 66, 52, 74,
+			49, 57, 80, 55, 0, 36, 53, 138, 45, 79, 98, 52, 82, 58, 56, 84, 52, 62, 91, 58, 0, 37,
+			64, 129, 48, 88, 88, 54, 90, 54, 60, 96, 55, 66, 104, 62, 0, 40, 75, 120, 50, 96, 74,
+			55, 99, 51, 62, 106, 57, 68, 113, 62, 0, 42, 89, 107, 50, 104, 60, 57, 108, 49, 64,
+			114, 56, 71, 123, 65, 0,
+		];
+
+		let before = raw.clone();
+		let image = read_cf_dibv5(&mut raw).unwrap();
+
+		// Not expecting any header fiddling to happen here. This is a bitmap in 24-bit format, with a header
+		// that says as much
+		assert_eq!(raw, before);
+
+		assert_eq!(image.width, 5);
+		assert_eq!(image.height, 5);
+
+		const EXPECTED: &[u8] = &[
+			107, 89, 42, 255, 60, 104, 50, 255, 49, 108, 57, 255, 56, 114, 64, 255, 65, 123, 71,
+			255, 120, 75, 40, 255, 74, 96, 50, 255, 51, 99, 55, 255, 57, 106, 62, 255, 62, 113, 68,
+			255, 129, 64, 37, 255, 88, 88, 48, 255, 54, 90, 54, 255, 55, 96, 60, 255, 62, 104, 66,
+			255, 138, 53, 36, 255, 98, 79, 45, 255, 58, 82, 52, 255, 52, 84, 56, 255, 58, 91, 62,
+			255, 144, 47, 36, 255, 110, 68, 42, 255, 66, 74, 48, 255, 49, 74, 52, 255, 55, 80, 57,
+			255,
+		];
+		assert_eq!(image.bytes, EXPECTED);
+	}
+
+	#[test]
+	fn chrome_dibv5() {
+		// A 5x5 sample of https://commons.wikimedia.org/wiki/File:PNG_transparency_demonstration_1.png
+		// (interestingly, the same sample as in the Firefox test - despite the pixel data being
+		// materially different!)
+		let mut raw = vec![
+			124, 0, 0, 0, 5, 0, 0, 0, 5, 0, 0, 0, 1, 0, 32, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 255,
+			32, 110, 105, 87, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
+			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 0,
+			0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 32, 38, 145, 192, 38, 65, 111, 158, 46, 73, 68,
+			107, 50, 73, 50, 92, 55, 79, 55, 100, 31, 46, 139, 190, 41, 76, 100, 152, 49, 81, 60,
+			110, 53, 83, 53, 108, 60, 91, 60, 118, 32, 59, 131, 187, 44, 86, 89, 150, 51, 89, 56,
+			121, 57, 95, 57, 127, 63, 103, 63, 139, 35, 71, 122, 186, 46, 95, 76, 150, 52, 99, 54,
+			136, 59, 105, 59, 146, 65, 113, 65, 156, 37, 86, 109, 184, 46, 103, 63, 155, 52, 107,
+			53, 152, 60, 114, 60, 162, 68, 123, 68, 174,
+		];
+
+		let before = raw.clone();
+		let image = read_cf_dibv5(&mut raw).unwrap();
+
+		// Chrome's header is dodgy. Expect that we fiddled with it.
+		assert_ne!(raw, before);
+
+		assert_eq!(image.width, 5);
+		assert_eq!(image.height, 5);
+
+		const EXPECTED: &[u8] = &[
+			109, 86, 37, 184, 63, 103, 46, 155, 53, 107, 52, 152, 60, 114, 60, 162, 68, 123, 68,
+			174, 122, 71, 35, 186, 76, 95, 46, 150, 54, 99, 52, 136, 59, 105, 59, 146, 65, 113, 65,
+			156, 131, 59, 32, 187, 89, 86, 44, 150, 56, 89, 51, 121, 57, 95, 57, 127, 63, 103, 63,
+			139, 139, 46, 31, 190, 100, 76, 41, 152, 60, 81, 49, 110, 53, 83, 53, 108, 60, 91, 60,
+			118, 145, 38, 32, 192, 111, 65, 38, 158, 68, 73, 46, 107, 50, 73, 50, 92, 55, 79, 55,
+			100,
+		];
+		assert_eq!(image.bytes, EXPECTED);
+	}
+}
+
+unsafe fn global_alloc(bytes: usize) -> Result<HGLOBAL, Error> {
+	let hdata = GlobalAlloc(GHND, bytes);
+	if hdata.is_null() {
+		Err(last_error("Could not allocate global memory object"))
+	} else {
+		Ok(hdata)
+	}
+}
+
+unsafe fn global_lock(hmem: HGLOBAL) -> Result<*mut u8, Error> {
+	let data_ptr = GlobalLock(hmem).cast::<u8>();
+	if data_ptr.is_null() {
+		Err(last_error("Could not lock the global memory object"))
+	} else {
+		Ok(data_ptr)
+	}
+}
+
+unsafe fn global_unlock_checked(hdata: HGLOBAL) {
+	// If the memory object is unlocked after decrementing the lock count, the function
+	// returns zero and GetLastError returns NO_ERROR. If it fails, the return value is
+	// zero and GetLastError returns a value other than NO_ERROR.
+	if GlobalUnlock(hdata) == 0 {
+		let err = io::Error::last_os_error();
+		if err.raw_os_error() != Some(0) {
+			log::error!("Failed calling GlobalUnlock when writing data: {}", err);
+		}
+	}
+}
+
+fn last_error(message: &str) -> Error {
+	let os_error = io::Error::last_os_error();
+	Error::unknown(format!("{message}: {os_error}"))
+}
+
+/// An abstraction trait over the different ways a Win32 function may return
+/// a value with a failure marker.
+///
+/// This trait helps unify error handling across varying `windows-sys` versions,
+/// providing a consistent interface for representing NULL values.
+trait ResultValue: Sized {
+	const NULL: Self;
+	fn failure(self) -> bool;
+}
+
+// windows-sys >= 0.59
+impl<T> ResultValue for *mut T {
+	const NULL: Self = core::ptr::null_mut();
+	fn failure(self) -> bool {
+		self == Self::NULL
+	}
+}
+
+// `windows-sys` 0.52
+impl ResultValue for isize {
+	const NULL: Self = 0;
+	fn failure(self) -> bool {
+		self == Self::NULL
+	}
 }
 
 /// A shim clipboard type that can have operations performed with it, but
@@ -517,7 +536,7 @@ impl Clipboard {
 		Ok(Self(()))
 	}
 
-	fn open(&mut self) -> Result<OpenClipboard, Error> {
+	fn open(&mut self) -> Result<OpenClipboard<'_>, Error> {
 		// Attempt to open the clipboard multiple times. On Windows, its common for something else to temporarily
 		// be using it during attempts.
 		//
@@ -573,36 +592,20 @@ impl<'clipboard> Get<'clipboard> {
 			return Err(Error::ContentNotAvailable);
 		}
 
-		let text_size = clipboard_win::raw::size(FORMAT)
-			.ok_or_else(|| Error::unknown("failed to read clipboard text size"))?;
-
-		// Allocate the specific number of WTF-16 characters we need to receive.
-		// This division is always accurate because Windows uses 16-bit characters.
-		let mut out: Vec<u16> = vec![0u16; text_size.get() / 2];
-
-		let bytes_read = {
-			// SAFETY: The source slice has a greater alignment than the resulting one.
-			let out: &mut [u8] =
-				unsafe { std::slice::from_raw_parts_mut(out.as_mut_ptr().cast(), out.len() * 2) };
-
-			let mut bytes_read = clipboard_win::raw::get(FORMAT, out)
-				.map_err(|_| Error::unknown("failed to read clipboard string"))?;
-
-			// Convert the number of bytes read to the number of `u16`s
-			bytes_read /= 2;
-
-			// Remove the NUL terminator, if it existed.
-			if let Some(last) = out.last().copied() {
-				if last == 0 {
-					bytes_read -= 1;
-				}
-			}
-
-			bytes_read
-		};
-
-		// Create a UTF-8 string from WTF-16 data, if it was valid.
-		String::from_utf16(&out[..bytes_read]).map_err(|_| Error::ConversionFailure)
+		// NB: Its important that whatever functionality decodes the text buffer from the clipboard
+		// uses `WideCharToMultiByte` with `CP_UTF8` (or an equivalent) in order to handle when both "text"
+		// and a locale identifier were placed on the clipboard. It is probable this occurs when an application
+		// is running with a codepage that isn't the current system's, such as under a locale emulator.
+		//
+		// In these cases, Windows decodes the text buffer with whatever codepage that identifier is for
+		// when creating the `CF_UNICODETEXT` buffer. Therefore, the buffer could then be in any format,
+		// not nessecarily wide UTF-16. We need to then undo that, taking the wide data and mapping it into
+		// the UTF-8 space as best as possible.
+		//
+		// (locale-specific text data, locale id) -> app -> system -> arboard (locale-specific text data) -> UTF-8
+		let mut out = Vec::new();
+		clipboard_win::raw::get_string(&mut out).map_err(|_| Error::ContentNotAvailable)?;
+		String::from_utf8(out).map_err(|_| Error::ConversionFailure)
 	}
 
 	pub(crate) fn html(self) -> Result<String, Error> {
@@ -620,20 +623,24 @@ impl<'clipboard> Get<'clipboard> {
 
 	#[cfg(feature = "image-data")]
 	pub(crate) fn image(self) -> Result<ImageData<'static>, Error> {
-		const FORMAT: u32 = clipboard_win::formats::CF_DIBV5;
-
 		let _clipboard_assertion = self.clipboard?;
+		let mut data = Vec::new();
 
-		if !clipboard_win::is_format_avail(FORMAT) {
+		let png_format: Option<u32> = clipboard_win::register_format("PNG").map(From::from);
+		if let Some(id) = png_format.filter(|&id| clipboard_win::is_format_avail(id)) {
+			// Looks like PNG is available! Let's try it
+			clipboard_win::raw::get_vec(id, &mut data)
+				.map_err(|_| Error::unknown("failed to read clipboard PNG data"))?;
+			return image_data::read_png(&data);
+		}
+
+		if !clipboard_win::is_format_avail(clipboard_win::formats::CF_DIBV5) {
 			return Err(Error::ContentNotAvailable);
 		}
 
-		let mut data = Vec::new();
-
-		clipboard_win::raw::get_vec(FORMAT, &mut data)
+		clipboard_win::raw::get_vec(clipboard_win::formats::CF_DIBV5, &mut data)
 			.map_err(|_| Error::unknown("failed to read clipboard image data"))?;
-
-		image_data::read_cf_dibv5(&data)
+		image_data::read_cf_dibv5(&mut data)
 	}
 
 	pub(crate) fn file_list(self) -> Result<Vec<PathBuf>, Error> {
@@ -718,6 +725,74 @@ impl<'clipboard> Set<'clipboard> {
 		image_data::add_cf_dibv5(open_clipboard, image)?;
 		Ok(())
 	}
+
+	pub(crate) fn file_list(self, file_list: &[impl AsRef<Path>]) -> Result<(), Error> {
+		const DROPFILES_HEADER_SIZE: usize = std::mem::size_of::<DROPFILES>();
+
+		let clipboard_assertion = self.clipboard?;
+
+		// https://learn.microsoft.com/en-us/windows/win32/shell/clipboard#cf_hdrop
+		// CF_HDROP consists of an STGMEDIUM structure that contains a global memory object.
+		// The structure's hGlobal member points to the resulting data:
+		// | DROPFILES | FILENAME | NULL | ... | nth FILENAME | NULL | NULL |
+		let dropfiles = DROPFILES {
+			pFiles: DROPFILES_HEADER_SIZE as u32,
+			pt: POINT { x: 0, y: 0 },
+			fNC: 0,
+			fWide: 1,
+		};
+
+		let mut data_len = DROPFILES_HEADER_SIZE;
+
+		let paths: Vec<_> = file_list
+			.iter()
+			.filter_map(|path| {
+				to_final_path_wide(path.as_ref()).map(|wide| {
+					// Windows uses wchar_t which is 16 bit
+					data_len += wide.len() * std::mem::size_of::<u16>();
+					wide
+				})
+			})
+			.collect();
+
+		if paths.is_empty() {
+			return Err(Error::ConversionFailure);
+		}
+
+		// Add space for the final null character
+		data_len += std::mem::size_of::<u16>();
+
+		unsafe {
+			let h_global = global_alloc(data_len)?;
+			let data_ptr = global_lock(h_global)?;
+
+			(data_ptr as *mut DROPFILES).write(dropfiles);
+
+			let mut ptr = data_ptr.add(DROPFILES_HEADER_SIZE) as *mut u16;
+
+			for wide_path in paths {
+				std::ptr::copy_nonoverlapping::<u16>(wide_path.as_ptr(), ptr, wide_path.len());
+				ptr = ptr.add(wide_path.len());
+			}
+
+			// Write final null character
+			ptr.write(0);
+
+			global_unlock_checked(h_global);
+
+			if SetClipboardData(CF_HDROP.into(), h_global as HANDLE).failure() {
+				GlobalFree(h_global);
+				return Err(last_error("SetClipboardData failed with error"));
+			}
+		}
+
+		add_clipboard_exclusions(
+			clipboard_assertion,
+			self.exclude_from_monitoring,
+			self.exclude_from_cloud,
+			self.exclude_from_history,
+		)
+	}
 }
 
 fn add_clipboard_exclusions(
@@ -839,18 +914,105 @@ fn wrap_html(ctn: &str) -> String {
 	let n_end_frag = n_start_frag + ctn.len();
 	let n_end_html = n_end_frag + c_end_frag.len();
 	format!(
-		"{}{}{:010}{}{:010}{}{:010}{}{:010}{}{}{}",
-		h_version,
-		h_start_html,
-		n_start_html,
-		h_end_html,
-		n_end_html,
-		h_start_frag,
-		n_start_frag,
-		h_end_frag,
-		n_end_frag,
-		c_start_frag,
-		ctn,
-		c_end_frag,
+		"{h_version}{h_start_html}{n_start_html:010}{h_end_html}{n_end_html:010}{h_start_frag}{n_start_frag:010}{h_end_frag}{n_end_frag:010}{c_start_frag}{ctn}{c_end_frag}"
 	)
 }
+
+/// Given a file path attempt to open it and call GetFinalPathNameByHandleW,
+/// on success return the final path as a NULL terminated u16 Vec
+fn to_final_path_wide(p: &Path) -> Option<Vec<u16>> {
+	let file = std::fs::OpenOptions::new()
+		// No read or write permissions are necessary
+		.access_mode(0)
+		// This flag is so we can open directories too
+		.custom_flags(FILE_FLAG_BACKUP_SEMANTICS)
+		.open(p)
+		.ok()?;
+
+	fill_utf16_buf(
+		|buf, sz| unsafe {
+			GetFinalPathNameByHandleW(file.as_raw_handle() as HANDLE, buf, sz, VOLUME_NAME_DOS)
+		},
+		|buf| {
+			let mut wide = Vec::with_capacity(buf.len() + 1);
+			wide.extend_from_slice(buf);
+			wide.push(0);
+
+			let hr = unsafe { PathCchStripPrefix(wide.as_mut_ptr(), wide.len()) };
+			// On success truncate invalid data
+			if hr == S_OK {
+				if let Some(end) = wide.iter().position(|c| *c == 0) {
+					// Retain NULL character
+					wide.truncate(end + 1)
+				}
+			}
+			wide
+		},
+	)
+}
+
+/// <https://github.com/rust-lang/rust/blob/f34ba774c78ea32b7c40598b8ad23e75cdac42a6/library/std/src/sys/pal/windows/mod.rs#L211>
+fn fill_utf16_buf<F1, F2, T>(mut f1: F1, f2: F2) -> Option<T>
+where
+	F1: FnMut(*mut u16, u32) -> u32,
+	F2: FnOnce(&[u16]) -> T,
+{
+	// Start off with a stack buf but then spill over to the heap if we end up
+	// needing more space.
+	//
+	// This initial size also works around `GetFullPathNameW` returning
+	// incorrect size hints for some short paths:
+	// https://github.com/dylni/normpath/issues/5
+	let mut stack_buf: [std::mem::MaybeUninit<u16>; 512] = [std::mem::MaybeUninit::uninit(); 512];
+	let mut heap_buf: Vec<std::mem::MaybeUninit<u16>> = Vec::new();
+	unsafe {
+		let mut n = stack_buf.len();
+		loop {
+			let buf = if n <= stack_buf.len() {
+				&mut stack_buf[..]
+			} else {
+				let extra = n - heap_buf.len();
+				heap_buf.reserve(extra);
+				// We used `reserve` and not `reserve_exact`, so in theory we
+				// may have gotten more than requested. If so, we'd like to use
+				// it... so long as we won't cause overflow.
+				n = heap_buf.capacity().min(u32::MAX as usize);
+				// Safety: MaybeUninit<u16> does not need initialization
+				heap_buf.set_len(n);
+				&mut heap_buf[..]
+			};
+
+			// This function is typically called on windows API functions which
+			// will return the correct length of the string, but these functions
+			// also return the `0` on error. In some cases, however, the
+			// returned "correct length" may actually be 0!
+			//
+			// To handle this case we call `SetLastError` to reset it to 0 and
+			// then check it again if we get the "0 error value". If the "last
+			// error" is still 0 then we interpret it as a 0 length buffer and
+			// not an actual error.
+			windows_sys::Win32::Foundation::SetLastError(0);
+			let k = match f1(buf.as_mut_ptr().cast::<u16>(), n as u32) {
+				0 if GetLastError() == 0 => 0,
+				0 => return None,
+				n => n,
+			} as usize;
+			if k == n && GetLastError() == windows_sys::Win32::Foundation::ERROR_INSUFFICIENT_BUFFER
+			{
+				n = n.saturating_mul(2).min(u32::MAX as usize);
+			} else if k > n {
+				n = k;
+			} else if k == n {
+				// It is impossible to reach this point.
+				// On success, k is the returned string length excluding the null.
+				// On failure, k is the required buffer length including the null.
+				// Therefore k never equals n.
+				unreachable!();
+			} else {
+				// Safety: First `k` values are initialized.
+				let slice = std::slice::from_raw_parts(buf.as_ptr() as *const u16, k);
+				return Some(f2(slice));
+			}
+		}
+	}
+}
